/*

Freeform Builds Nixie Style Clock Arduino Nano Code
Visit www.freeformbuilds.com for a step by step guide on how to build the clock as well as a parts list

-----

The code is taken and modified from the following sources: 

 "$100 Pace and Wall Clock by Michael Jacobson" - https://www.instructables.com/100-Pace-Clock-or-Wall-Clock/
 "Controlling large 7-segment displays by Nathan Seidle /SparkFun Electronics" - https://learn.sparkfun.com/tutorials/large-digit-driver-hookup-guide#resources-and-going-further
 
-----

Here's how to hook up the Arduino nano pins to the Freeform Builds 7 seg PCB: 

Arduino pin 6 -> CLOCK
Arduino Pin 5 -> LATCH
Arduino Pin 7 -> SER_I
Arduino 5V -> 5V 
GND -> GND 

The Freeform Builds 7 seg PCB also requires a voltage source based on the LEDs being used
in the case of the guide on the website I am using a 5V 2A power source to directly power the modules. The power source needs to be connected directly to V+ on the board. 

-----

To connect the DS3231 mini RTC to the Arduino nano do the following: 

Arduino pin A5 -> SCL
A4 -> SDA
3.3V -> 3V3
GND -> GND
  
*/
#include <Wire.h>
#define DS3231_I2C_ADDRESS 104
byte tMSB, tLSB;
float temp3231;
byte seconds, minutes, hours, day, date, month, year;
char weekDay[4];
bool flip_3rd_digit = false; // set true if you want to flip it so the colon looks like a clock

//GPIO declarations
//-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
byte segmentClock = 6;
byte segmentLatch = 5;
byte segmentData = 7;

//-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

void setup()
{
  Serial.begin(9600);
  Serial.println("Freeform Builds Nixie Style Clock");

  // RTC 
  Wire.begin();
  Wire.beginTransmission(0x68); // address DS3231
  Wire.write(0x0E); // select register
  Wire.write(0b00011100); // write register bitmap, bit 7 is /EOSC
  Wire.endTransmission();

  // LED Driver
  pinMode(segmentClock, OUTPUT);
  pinMode(segmentData, OUTPUT);
  pinMode(segmentLatch, OUTPUT);
  digitalWrite(segmentClock, LOW);
  digitalWrite(segmentData, LOW);
  digitalWrite(segmentLatch, LOW);

}

// main loop - Display is updated only with the time changes (1 per sec or min)
void loop()
{
   int currtime;  // stores in the current display time
   int oldtime=0; 
   digitalWrite(8,LOW);  // pull output low on Pin 8 and if connected to Pin 9 will put in Wall Clock mode
   showNumber(8888);  // test all the segments
   delay(1500);
   while(1) {
     watchConsole();
     get3231Date();
      {             
       currtime = hours*100+minutes;
     }  
     if (currtime != oldtime) {
        showNumber(currtime);
        oldtime = currtime;
        Serial.print(weekDay); Serial.print(", "); Serial.print(month, DEC); Serial.print("/"); Serial.print(date, DEC); Serial.print("/"); Serial.print(year, DEC); Serial.print(" - ");
        Serial.print(hours, DEC); Serial.print(":"); Serial.print(minutes, DEC); Serial.print(":"); Serial.print(seconds, DEC);
        Serial.print(" - Temp: "); Serial.println(get3231Temp());
     
     } else {
        delay(400);  // need to delay or serial input is not handled
     }
   }
}

//Takes a 4 digit number and outputs to the driver board (one at a time).
void showNumber(float value)
{
  int number = abs(value); //Remove negative signs and any decimals

  for (byte x = 0 ; x < 4 ; x++)
  {
    int remainder = number % 10;
    switch (x)
    {
      // postNumber digit decimal_pt flip
      case 1: postNumber(remainder, true, flip_3rd_digit); break; // 3rd digit
      case 2: postNumber(remainder, true, false); break; // 2nd digit
      default: postNumber(remainder, false, false); // 1st and 4th digits
    }
    number /= 10;
  }

  //Latch the current segment data
  digitalWrite(segmentLatch, LOW);
  digitalWrite(segmentLatch, HIGH); //Register moves storage register on the rising edge of RCK
}

//Given a number, or '-', shifts it out to the display.  It can also flip it and add a decimal point
void postNumber(byte number, boolean decimal, boolean flip)
{
  //    -  A
  //   / / F/B
  //    -  G
  //   / / E/C
  //    -. D/DP

#define a  1<<0
#define b  1<<6
#define c  1<<5
#define d  1<<4
#define e  1<<3
#define f  1<<1
#define g  1<<2
#define dp 1<<7

  byte segments;
  if (flip) {
    switch (number)
    {
      case 1: segments = e | f; break;
      case 2: segments = a | b | d | e | g; break;
      case 3: segments = a | d | e | f | g; break;
      case 4: segments = c | e | f | g; break;
      case 5: segments = a | c | d | f | g; break;
      case 6: segments = a | b | c | d | f | g; break;
      case 7: segments = d | e | f; break;
      case 8: segments = a | b | c | d | e | f | g; break;
      case 9: segments = a | g | c | d | e | f; break;
      case 0: segments = a | c | d | e | f | b; break;
      case ' ': segments = 0; break;
      case 'c': segments = g | e | d; break;
      case '-': segments = g; break;
    }
  } else {
    switch (number)
    {
      case 1: segments = b | c; break;
      case 2: segments = a | b | d | e | g; break;
      case 3: segments = a | b | c | d | g; break;
      case 4: segments = f | g | b | c; break;
      case 5: segments = a | f | g | c | d; break;
      case 6: segments = a | f | g | e | c | d; break;
      case 7: segments = a | b | c; break;
      case 8: segments = a | b | c | d | e | f | g; break;
      case 9: segments = a | b | c | d | f | g; break;
      case 0: segments = a | b | c | d | e | f; break;
      case ' ': segments = 0; break;
      case 'c': segments = g | e | d; break;
      case '-': segments = g; break;
    }
  }
  if (decimal) segments |= dp;

  //Clock these bits out to the drivers
  for (byte x = 0 ; x < 8 ; x++)
  {
    digitalWrite(segmentClock, LOW);
    digitalWrite(segmentData, segments & 1 << (7 - x));
    digitalWrite(segmentClock, HIGH); //Data transfers to the register on the rising edge of SRCK
  }
}

// sets the RTC via the serial port
void set3231Date()
{
//T(sec)(min)(hour)(dayOfWeek)(dayOfMonth)(month)(year)
//T(00-59)(00-59)(00-23)(1-7)(01-31)(01-12)(00-99)
//Example: 02-Feb-09 @ 19:57:11 for the 3rd day of the week -> T1157193020209
// In TCL -> clock format [clock seconds] -format "T%S%M%H%u%d%m%g"
  seconds = (byte) ((Serial.read() - 48) * 10 + (Serial.read() - 48)); // Use of (byte) type casting and ascii math to achieve result.  
  minutes = (byte) ((Serial.read() - 48) *10 +  (Serial.read() - 48));
  hours   = (byte) ((Serial.read() - 48) *10 +  (Serial.read() - 48));
  day     = (byte) (Serial.read() - 48);
  date    = (byte) ((Serial.read() - 48) *10 +  (Serial.read() - 48));
  month   = (byte) ((Serial.read() - 48) *10 +  (Serial.read() - 48));
  year    = (byte) ((Serial.read() - 48) *10 +  (Serial.read() - 48));
  Wire.beginTransmission(DS3231_I2C_ADDRESS);
  Wire.write(0x00);
  Wire.write(decToBcd(seconds));
  Wire.write(decToBcd(minutes));
  Wire.write(decToBcd(hours));
  Wire.write(decToBcd(day));
  Wire.write(decToBcd(date));
  Wire.write(decToBcd(month));
  Wire.write(decToBcd(year));
  Wire.endTransmission();
}

void get3231Date()
{
  // send request to receive data starting at register 0
  Wire.beginTransmission(DS3231_I2C_ADDRESS); // 104 is DS3231 device address
  Wire.write(0x00); // start at register 0
  Wire.endTransmission();
  Wire.requestFrom(DS3231_I2C_ADDRESS, 7); // request seven bytes

  if(Wire.available()) {
    seconds = Wire.read(); // get seconds
    minutes = Wire.read(); // get minutes
    hours   = Wire.read();   // get hours
    day     = Wire.read();
    date    = Wire.read();
    month   = Wire.read(); //temp month
    year    = Wire.read();
       
    seconds = (((seconds & B11110000)>>4)*10 + (seconds & B00001111)); // convert BCD to decimal
    minutes = (((minutes & B11110000)>>4)*10 + (minutes & B00001111)); // convert BCD to decimal
    hours   = (((hours & B00110000)>>4)*10 + (hours & B00001111)); // convert BCD to decimal (assume 24 hour mode)
    day     = (day & B00000111); // 1-7
    date    = (((date & B00110000)>>4)*10 + (date & B00001111)); // 1-31
    month   = (((month & B00010000)>>4)*10 + (month & B00001111)); //msb7 is century overflow
    year    = (((year & B11110000)>>4)*10 + (year & B00001111));
  }
  else {
    //oh noes, no data!
  }
 
  switch (day) {
    case 1:
      strcpy(weekDay, "Sun");
      break;
    case 2:
      strcpy(weekDay, "Mon");
      break;
    case 3:
      strcpy(weekDay, "Tue");
      break;
    case 4:
      strcpy(weekDay, "Wed");
      break;
    case 5:
      strcpy(weekDay, "Thu");
      break;
    case 6:
      strcpy(weekDay, "Fri");
      break;
    case 7:
      strcpy(weekDay, "Sat");
      break;
  }
}

float get3231Temp()
{
  //temp registers (11h-12h) get updated automatically every 64s
  Wire.beginTransmission(DS3231_I2C_ADDRESS);
  Wire.write(0x11);
  Wire.endTransmission();
  Wire.requestFrom(DS3231_I2C_ADDRESS, 2);
 
  if(Wire.available()) {
    tMSB = Wire.read(); //2's complement int portion
    tLSB = Wire.read(); //fraction portion
   
    temp3231 = (tMSB & B01111111); //do 2's math on Tmsb
    temp3231 += ( (tLSB >> 6) * 0.25 ); //only care about bits 7 & 8
  }
  else {
    //oh noes, no data!
  }
   
  return temp3231;
}


void watchConsole()
{
 if (Serial.available()) {      // Look for char in serial queue and process if found
   if (Serial.read() == 84) {      //If command = "T" Set Date
     set3231Date();
     get3231Date();
     Serial.println("Clock Update");
   }
 }
}

byte decToBcd(byte val)
{
 return ( (val/10*16) + (val%10) );
}